-- Various genesis-related scripts (misc utils, logger etc)
-- Stateless - nothing to initialize here
utils = {}

mud = {}
mud.state = {}

log = {
    last_level = 0,
    level_others = 10, -- for those that are not configured
}

--------------------------------------------------------- Help
genesis_help = genesis_help or {}

genesis_help._modules = genesis_help._modules or {}
genesis_help._briefs = genesis_help._briefs or {}
genesis_help._topics = genesis_help._topics or {}
genesis_help._apis = genesis_help._apis or {}


local function ensureHelpExists(mod)
    genesis_help._modules[mod] = genesis_help._modules[mod] or ""
    genesis_help._briefs[mod] = genesis_help._briefs[mod] or ""
    genesis_help._topics[mod] = genesis_help._topics[mod] or {}
    genesis_help._apis[mod] = genesis_help._apis[mod] or {}
end

function utils.moduleExists(name)
    return genesis_help._modules[name] ~= nil
end

function utils.setHelpModule(mod, brief, help)
    ensureHelpExists(mod)
    genesis_help._modules[mod] = help
    genesis_help._briefs[mod] = brief
end

function utils.setHelpModuleTopic(mod, topic, help)
   ensureHelpExists(mod)
   genesis_help._topics[mod][topic] = help
end

function utils.setHelpModuleApi(mod, topic, help)
    ensureHelpExists(mod)
    genesis_help._apis[mod][topic] = help
end

function utils.getHelpIndex()
    local idx = {"Genesis module help system. Usage: ",
                 "ghelp                  - list available modules (and this page)",
                 "ghelp <module>         - help index for module",
                 "ghelp <module> <topic> - help on specific topic in module",
                 "gapi <module>          - API index for module",
                 "gapi <module> <fcn>    - documentation for function in module",
                 "\nThe following modules are available:"}

    for k,v in pairs(genesis_help._modules) do
        if genesis_help._briefs[k] ~= "" then
            table.insert(idx, k.." - "..genesis_help._briefs[k])
        else
            table.insert(idx, k)
        end
    end
    return table.concat(idx, "\n")
end

function utils.getHelpModuleIndex(mod)
    if not utils.moduleExists(mod) then
        return "Module "..mod.." not found.\n"
    end

    ensureHelpExists(mod)

    local idx = {
        "Module "..mod.." - "..genesis_help._briefs[mod],
        genesis_help._modules[mod],
        "Available topics: "
    }
    for topic,val in pairs(genesis_help._topics[mod]) do
        table.insert(idx, topic)
    end
    return table.concat(idx, "\n")
end

function utils.getHelpModuleTopic(mod, topic)
    if not utils.moduleExists(mod) then
        return "Module "..mod.." not found.\n"
    end

    ensureHelpExists(mod)
    if genesis_help._topics[mod][topic] then
        return string.format("%s: \n%s\n", topic, genesis_help._topics[mod][topic])
    else
        return "Topic ".. topic .. " not found in mod "..mod.."\n"
    end
end

function utils.getHelpModuleApiIndex(mod)
    if not utils.moduleExists(mod) then
        return "Module "..mod.." not found.\n"
    end

    ensureHelpExists(mod)

    local idx = {
        "Module "..mod.." - "..genesis_help._briefs[mod],
        genesis_help._modules[mod],
        "Available APIs: "
    }
    for topic,val in pairs(genesis_help._apis[mod]) do
        table.insert(idx, topic)
    end
    return table.concat(idx, "\n")
end

function utils.getHelpModuleApi(mod, topic)
    if not utils.moduleExists(mod) then
        return "Module "..mod.." not found.\n"
    end

    ensureHelpExists(mod)
    if genesis_help._apis[mod][topic] then
        return genesis_help._apis[mod][topic]
    else
        return "API ".. topic .. " not found in mod "..mod.."\n"
    end
end

utils.setHelpModule("genesis", "Core genesis functions.",
"Genesis module is base for all the other modules.TODO: fill.")
--------------------------------------------------------- Mud related stuff

utils.setHelpModuleApi("genesis", "sendCmd", "mud.sendCmd(cmd, silent) - wrapper tracking last posted command.")
function mud.sendCmd(cmd, silent)
    silent = silent or false
    mud.state.last_command = cmd
    send(cmd, silent)
end

utils.setHelpModuleApi("genesis", "sendMove", "mud.sendMove(dir, force) - sends move to game. May be masked by mapper.")
function mud.sendMove(dir, force, speedwalk)
    force = force or false
    mud.state.last_movement = dir
    if mud.sendMoveOverride then
        mud.sendMoveOverride(dir, force, speedwalk)
    else
        cecho("<steel_blue>Moving "..dir.."\n")
        mud.sendCmd(dir, true)
    end
end


-------------------------------------------------------- Number stuff

-- Expands list in format (a|an|a pair of) something, ... and something
-- into list of single items
-- TODO: optimize
-- TODO: doesn't handle "and" items - not sure how to handle yet
function utils.expandList(line)
    local items = string.split(line, ", ")
    local items_parsed = {}

    for i, what in pairs(items) do
    -- not very effective - todo optimize
        what = what:gsub("^a pair of ","")
        what = what:gsub("^an ","")
        what = what:gsub("^a ","")
        if what:find(" and ") then
            what1 = what:sub(1, what:find(" and ") - 1)
            what2 = what:sub(what:find(" and ") + 5)
            what2 = what2:gsub("^a pair of ","")
            what2 = what2:gsub("^an ","")
            what2 = what2:gsub("^a ","")
            table.insert(items_parsed, what1)
            table.insert(items_parsed, what2)
        else
            table.insert(items_parsed, what)
        end
    end
    return items_parsed
end

function utils.doMany(num, what)
    if num < 1 then
     cecho("<red>Usage: for <num> <action|action %s>")
     return
    end

    -- with ordinal
    if string.find(what, "%s") ~= nil then
      for i=1,num do
          expandAlias(string.format(what, utils.numToOrd(i)))
      end
    else
      for i=1,num do
          expandAlias(what)
      end
    end
end


function utils.reverseTable(table)
 table_out = {}
 for k,v in pairs(table) do
  table_out[v] = k
 end
 return table_out
end

function utils.cloneTable(orig)
    local orig_type = type(orig)
    local copy
    if orig_type == 'table' then
        copy = {}
        for orig_key, orig_value in next, orig, nil do
            copy[utils.cloneTable(orig_key)] = utils.cloneTable(orig_value)
        end
        setmetatable(copy, utils.cloneTable(getmetatable(orig)))
    else -- number, string, boolean, etc
        copy = orig
    end
    return copy
end


function utils.numToDigit(num)
 local i = table.index_of(_num, num)
 if i == nil then
  gLogError("gNumToDigit: number not found:"..num.."\n")
  return nil
 end
 return i - 1
end

function utils.digitToNum(n)
 if n >= 0 and n < #_num then
  return _num[n + 1]
 end
 return nil
end

function utils.ordToDigit(ord)
 local i = table.index_of(_ord, ord)
 if i == nil then
  gLogError("gNumToDigit: number not found:"..ord.."\n")
  return nil
 end
 return i
end

function utils.digitToOrd(n)
 if n > 0 and n < #_num then
  return _ord[n]
 end
 return nil
end


function utils._numToWord(num, ones, tens)
    if (num < 1) then
        return "zero";
    elseif (num < 20) then
        return ones[num]
    elseif (num < 100) then
        local ret = tens[math.floor(num / 10) - 1]
        if (num % 10) > 0 then
            ret = ret.."-"..ones[num % 10]
        end
        return ret
    elseif (num < 1000) then
        local ret = utils._numToWord(math.floor(num / 100), ones, tens) .. " hundred"
        if (num % 100) > 0 then
            ret = ret .. " "..utils._numToWord(num % 100, ones, tens)
        end
        return ret
    elseif (num < 10000) then
        local ret = utils._numToWord(math.floor(num / 1000), ones, tens) .. " thousand"
        if (num % 1000) > 0 then
            ret = ret .. " "..utils._numToWord(num % 1000, ones, tens)
        end
        return ret
    end
    return "UnknownNumber"
end


function utils.numToWord(num)
    return utils._numToWord(num, utils.num, utils.numt)
end

function utils._numToOrd(num, ones, tens)
    if (num < 1) then
        return "zero";
    elseif (num < 20) then
        return ones[num]
    elseif (num < 100) then
        local ret = tens[math.floor(num / 10) - 1]
        if (num % 10) > 0 then
            ret = ret.."y-"..ones[num % 10]
        else
            ret = ret .. "ieth"
        end
        return ret
    elseif (num == 100) then
        return "hundredth"
    else
        return "UnknownNumber"
    end
end


function utils.numToOrd(num)
    return utils._numToOrd(num, utils.ord, utils.ordt)
end


-- Copied from genesis gamelib
utils.num = { "one", "two", "three", "four", "five", "six", "seven",
          "eight", "nine", "ten","eleven","twelve","thirteen",
          "fourteen", "fifteen", "sixteen", "seventeen", "eighteen",
          "nineteen"};
utils.ord = { "first", "second", "third", "fourth", "fifth", "sixth",
          "seventh", "eighth", "ninth", "tenth","eleventh","twelfth",
          "thirteenth", "fourteenth", "fifteenth", "sixteenth",
          "seventeenth", "eighteenth", "nineteenth"};
utils.numt = { "twenty", "thirty", "forty", "fifty", "sixty", "seventy",
          "eighty", "ninety"};
utils.ordt = { "twent", "thirt", "fort", "fift", "sixt", "sevent", "eight",
           "ninet" };


------------------------------------------------------------ LOGGER
function log.log(level, mod, submodule, fmt,...)
    local name

    if not submodule then
        submodule = ""
    end

    if mod and mod.settings and mod.settings.name then
        name = mod.settings.name
    else
        name = "??"
    end

    local lr = log.level_others

    if mod and mod.settings and mod.settings.debug_levels and mod.settings.debug_levels[submodule] then
        lr = mod.settings.debug_levels[submodule]
    end

    if submodule ~="" then
        submodule = "-"..submodule
    end

    log.last_level = lr

    if level <= lr then
        local format = "<green>%s%s: "
        if level < 2 then
            format = format .. "<red:yellow>"
        elseif level < 3 then
            format = format .. "<red>"
        elseif level < 6 then
            format = format .. "<white>"
        else
            format = format .. "<dim_grey>"
        end

        cecho(string.format(format..fmt, name, submodule, ...))
    end
end

function log.always(mod, submodule, fmt, ...)
    if not submodule then
        submodule = ""
    else
        submodule = "-"..submodule
    end

    if mod and mod.settings and mod.settings.name then
        cecho(string.format("<green>%s%s: <white>"..fmt, mod.settings.name, submodule, ...))
    else
        cecho(string.format("<green>??%s: <white>"..fmt, submodule, ...))
    end
end

function log.alwaysCont(fmt, ...)
    cecho(string.format(fmt, ...))
end

function log.dbg(mod, submodule, fmt,...)
    return log.log(8, mod, submodule, fmt, ...)
end

function log.info(mod, submodule, fmt,...)
    return log.log(4, mod, submodule, fmt, ...)
end

function log.crit(mod, submodule, fmt, ...)
    return log.log(1, mod, submodule, fmt, ...)
end

function log.err(mod, submodule, fmt,...)
    return log.log(2, mod, submodule,  fmt, ...)
end

function log.cont(mod, submodule, fmt,...)
    local lr = log.level_others

    if mod and mod.settings and mod.settings.debug_levels and mod.settings.debug_levels[submodule] then
        lr = mod.settings.debug_levels[submodule]
    end

    if log.last_level <= lr then
        cecho(string.format(fmt, ...))
    end
end
