-- mapping utils
-- influenced by Jor'Mox's Generic Map Script
-- partially even copied from it ;)

local settings = tr.settings
local state = tr.state
local map = tr.state.mapper

-- sets area for newly created rooms (when unknown)
function tr.setArea(name)
    local at = getAreaTable()

    if at[name] then
        map.area_id = at[name]
    else
        map.area_id = addAreaName(name)
        if not map.area_id then
            log.err(tr, "mapper", "tr.setArea: Can't create new area.\n")
            return
        end
    end
    log.info(tr, "mapper", "Switched to area '%s' with id %d\n", name, map.area_id)
end


function tr.mapperDisable()
    if map.enabled then
        map.enabled = false
        log.info(tr, "mapper", "Mapper is disabled.\n")
        tr.updateRoomInfo()
    end
end

function tr.mapperEnable()
    if not map.enabled then
        map.enabled = true
        log.info(tr, "mapper", "Mapper is enabled in %s mode.\n", map.mode)
        tr.updateRoomInfo()
    end
end

function tr.mapperSetMode(mode)
    -- TODO: checks
    map.mode = mode
    log.info(tr, "mapper", "Set mapper mode to %s.\n", mode)
    tr.updateRoomInfo()
end

function tr.mapperCreateSpecialExit(command, dir)
    if not map.enabled then
        log.err(tr, "mapper", "Enable mapper to create special exits.\n")
        return
    end
    if state.move_ongoing then
        log.err(tr, "mapper", "Movement ongoing. Wait until it finishes.\n")
        return
    end

    if not dir then
        dir = "e"
    else
        tr.setRoomMeta("override_"..dir, command)
    end
    map.placement_hint = { tr.exitOfs_x[tr.dir[dir]],
                           tr.exitOfs_y[tr.dir[dir]],
                           tr.exitOfs_z[tr.dir[dir]] }
    mud.sendMove(command)
end

-- converts current unknown room to new one - good for places like ships
function tr.mapperCreateRoom()
    local room_id = state.present.room_id
    local area_for_room = map.area_id

    if room_id then
        log.err(tr, "mapper", "Room already exits.\n")
        return
    end

    if not area_for_room then
        log.err(tr, "mapper", "Define area for new room first.\n")
        return
    end

    room_id = createRoomID()
    setRoomIDbyHash(room_id, state.present.room_hash)
    addRoom(room_id)
    state.present.room_id = room_id
    state.present.area_id = area_for_room
    setRoomName(room_id, state.gmcp_room.short)
    setRoomArea(room_id, area_for_room)
    state.present.area_name = getRoomAreaName(state.present.area_id)
    log.info(tr, "mapper", "Created new room %d: %s\n", room_id, state.gmcp_room.short)

    -- position room
    setRoomCoordinates(room_id, 0, 0, 0)
    tr.updateRoomInfo()
    centerview(room_id)
end

function tr.updateRoomStubs(room_added)
    local room_id = state.present.room_id

    if not room_id or not state.gmcp_room then
        return
    end

-- create exit stubs for gmcp-reported exits
    local exits = getRoomExits(room_id)
    local stubs = getExitStubs1(room_id)
    -- add stubs to exits
    if stubs then
        for k,v in ipairs(stubs) do
            exits[tr.dir_long[v]] = true
        end
    end
    for i,v in pairs(state.gmcp_room.exits) do
        local exit_linked = tr.dir[v] and exits[tr.dir_long[tr.dir[v]]]

        if tr.dir[v] then
            if not exit_linked then
                --log.log(10, tr, "mapper", "Setting stub to %s\n", v)
                setExitStub(room_id, tr.dir[v], true)
                -- New room, notify player
                if not room_added then
                    log.info(tr, "mapper", "<red>New exit at known room: %s!!!\n", v)
                end
            end
        end
    end
end

function tr.mapperMoved()
    local placement_hint = map.placement_hint
    if not map.enabled then
        tr.updateRoomStubs()
        map.placement_hint = nil
        return
    end

    local area_for_room = map.area_id

    if state.past.area_id then
        area_for_room = state.past.area_id
    end

    if not area_for_room then
        log.error(tr, "mapper", "Can't create room with unknown area.\n")
        return
    end

    -- do the right thing if we're on
    local present = state.present
    local past = state.past
    local room_id = present.room_id
    local past_id = past.room_id
    local room_added = false
    local dir = state.last_dir
    state.last_dir = nil


    if not present.room_hash then
        log.err(tr, "mapper", "Cant create room without GMCP id.\n")
        return
    end

    if not dir then
        log.err(tr, "mapper", "Unknown direction of last move. If it's special exit use 'sexit'.\n")
        return
    end

    -- room doesn't exist - must create new one
    if not room_id then
        room_id = createRoomID()
        setRoomIDbyHash(room_id, present.room_hash)
        addRoom(room_id)
        present.room_id = room_id
        present.area_id = area_for_room
        setRoomName(room_id, state.gmcp_room.short)
        setRoomArea(room_id, area_for_room)
        present.area_name = getRoomAreaName(present.area_id)
        log.info(tr, "mapper", "Created new room %d: %s\n", room_id, state.gmcp_room.short)
        room_added = true

        -- position room
        if not past_id then
                setRoomCoordinates(room_id, 0, 0, 0)
        else
            -- todo streching
            local x, y, z = getRoomCoordinates(past_id)

            if tr.dir[dir] then
                x = x + tr.exitOfs_x[tr.dir[dir]]
                y = y + tr.exitOfs_y[tr.dir[dir]]
                z = z + tr.exitOfs_z[tr.dir[dir]]
            elseif placement_hint then
                x = x + placement_hint[1]
                y = y + placement_hint[2]
                z = z + placement_hint[3]
            else
                log.info(tr, "mapper", "Non-standard direction detected, placing room randomly.\n")
                x = x + 1
            end
            map.placement_hint = nil
            setRoomCoordinates(room_id, x, y, z)
        end
    end

    if past_id then
    -- link past room
        if tr.dir[dir] then
            setExit(past_id, room_id, tr.dir[dir])
            -- backlink if there is exit in reverse direction
            if map.mode == "complex" or map.mode == "smart" then
                local rev = tr.dir[tr.dir_rev[dir]]
                if rev then -- should always be true
                    setExit(room_id, past_id, rev)
                end
            end
        else -- special exit
            addSpecialExit(past_id, room_id, dir)
        end
    end

    tr.updateRoomStubs(room_added)

    if room_added then
        tr.updateRoomInfo()
    end
    centerview(room_id)
end
