-- room management/state
local settings = tr.settings
local state = tr.state

function tr.setRoomMeta(key, value)
    local room = state.present.room_id

    if not room  then
        log.err(tr, "room", "Don't know that room, can't set metadata for it.\n")
        return
    end

    local key_ok = false

    if settings.room.meta_allowed[key] then
        key_ok = true
    else
        local u = string.find(key, "_")
        if u then
            local meta = string.sub(key, 1, u)
            local dir = string.sub(key, u + 1)

            key_ok = (settings.room.meta_allowed_exit[meta] ~=nil) and (tr.dir_rev[dir] ~=nil)

            -- Special handling of blocks
            if meta == "block_" then
                lockExit(room, dir, true)
            end
        end
    end
    if key_ok then
        log.info(tr, "room", "Setting room %d meta: %s = %s\n", room, tostring(key), tostring(value))
        setRoomUserData(room, tostring(key), tostring(value))
        tr.updateRoomInfo()
    else
        log.err(tr, "room", "Can't set meta %s for room - there is no such meta defined.\n", key)
    end
end

function tr.clearRoomMeta(key)
    local room = state.present.room_id

    if not room then
        log.err(tr, "room", "Don't know room, can't clear metadata.\n")
        return
    end

    log.info(tr, "room", "Clearing room %d meta. Was %s = %s\n", room, tostring(key), getRoomUserData(room, tostring(key)))
    clearRoomUserDataItem(room, tostring(key))

    local u = string.find(key, "_")
    if u then
        local meta = string.sub(key, 1, u)
        local dir = string.sub(key, u + 1)

        key_ok = (settings.room.meta_allowed_exit[meta] ~=nil) and (tr.dir_rev[dir] ~=nil)

        -- Special handling of blocks
        if meta == "block_" then
            lockExit(room, dir, false)
        end
    end

    tr.updateRoomInfo()
end

function tr.getRoomMeta(key)
    local room = state.present.room_id

    if not room then
        log.err(tr, "room", "Don't know room, can't query metadata.\n")
        return
    end

    return getRoomUserData(room, tostring(key))
end


function tr.getRoomInfo(full)
    local desc = {}

    if state.present.room_id and getRoomName(state.present.room_id) then
        local shown_exits = {}
        table.insert(desc,
                    string.format("<sea_green>Room <dim_grey>'%s'(%d): <sea_green>%s\n",
                                state.present.room_hash, state.present.room_id,
                                getRoomName(state.present.room_id)))
        table.insert(desc, "<sea_green>Exits: ")
        for i,dir in pairs(tr.state.gmcp_room.exits) do
            table.insert(desc, string.format("%s%s ", tr.briefExitColor(state.present.room_id, dir), dir))
            shown_exits[dir] = true
        end

        -- add invisible exits
        local exits = getRoomExits(state.present.room_id)
        for dir,no in pairs(exits) do
            if not shown_exits[dir] then
                table.insert(desc, string.format("%s[%s] ", tr.briefExitColor(state.present.room_id, tr.dirToShort(dir)), dir))
                shown_exits[dir] = true
            end
        end
        -- add special exits
        exits = getSpecialExitsSwap(state.present.room_id)
        for dir,no in pairs(exits) do
            if not shown_exits[dir] then
                table.insert(desc, string.format("%s[%s] ", tr.briefExitColor(state.present.room_id, tr.dirToShort(dir)), dir))
                shown_exits[dir] = true
            end
        end
        -- add remaining blocks and overrides
        for i, dir in pairs(tr.dir_short) do
            local long = tr.dirToLong(dir)
            --display(dir)
            if not shown_exits[long] then
                local short = dir
                local block = getRoomUserData(state.present.room_id, "block_"..short)
                local override = getRoomUserData(state.present.room_id, "override_"..short)

                if block ~= "" then
                    table.insert(desc, string.format("<red>{%s} ", long))
                elseif override ~="" then
                    table.insert(desc, string.format("<cyan>{%s} ", long))
                end
            end
        end
        table.insert(desc, "\n")
        local ud = getAllRoomUserData(tr.state.present.room_id)
        for i, v in pairs(ud) do
            if settings.room.meta_brief[i] then
                table.insert(desc, string.format("<%s>%s: %s\n", settings.room.meta_brief[i], i, v))
            elseif full then
                table.insert(desc, string.format("<sea_green>%s: %s\n", i, v))
            end
        end
    else
        table.insert(desc, string.format("<plum>Unknown room <dim_grey>'%s': <plum>%s\n<sea_green>Exits: <green>%s\n",
            state.present.room_hash, tr.state.gmcp_room.short,
            table.concat(tr.state.gmcp_room.exits, " ")))
    end
    return table.concat(desc, "")
end


-- display most important room information
function tr.updateRoomInfo()
    local info = tr.getRoomInfo()
    tr.ui.room_info:clear()
    tr.ui.room_info:cecho(info)
    if state.mapper.enabled then
        tr.ui.room_info:cecho("Mapper: "..state.mapper.mode.."\n")
    else
        tr.ui.room_info:cecho("Mapper: off\n")
    end
end
